@@ -1,5 +1,5 @@
 '''
-SimpleEval - (C) 2013-2016 Daniel Fairhead
+SimpleEval - (C) 2013-2017 Daniel Fairhead
 -------------------------------------
 
 An short, easy to use, safe and reasonably extensible expression evaluator.
@@ -38,8 +38,10 @@ Contributors:
 - corro (Robin Baumgartner) (py3k)
 - dratchkov (David R) (nested dicts)
 - marky1991 (Mark Young) (slicing)
-- T045T (Nils Berg) (!=, py3kstr, obj.attributes)
+- T045T (Nils Berg) (!=, py3kstr, obj.
 - perkinslr (Logan Perkins) (.__globals__ or .func_ breakouts)
+- impala2 (Kirill Stepanov) (massive _eval refactor)
+- gk (ugik) (Other iterables than str can DOS too, and can be made)
 
 -------------------------------------
 Usage:
@@ -117,9 +119,9 @@ class FunctionNotDefined(InvalidExpression):
 class NameNotDefined(InvalidExpression):
     ''' a name isn't defined. '''
     def __init__(self, name, expression):
+        self.name = name
         self.message = "'{0}' is not defined for expression '{1}'".format(
             name, expression)
-        self.name = name
         self.expression = expression
 
         # pylint: disable=bad-super-call
@@ -147,8 +149,8 @@ class NumberTooHigh(InvalidExpression):
     pass
 
 
-class StringTooLong(InvalidExpression):
-    ''' That string is **way** too long, baby. '''
+class IterableTooLong(InvalidExpression):
+    ''' That iterable is **way** too long, baby. '''
     pass
 
 
@@ -170,22 +172,22 @@ def safe_power(a, b):  # pylint: disable=invalid-name
 
 
 def safe_mult(a, b):  # pylint: disable=invalid-name
-    ''' limit the number of times a string can be repeated... '''
-    if isinstance(a, str) or isinstance(b, str):
-        if isinstance(a, int) and a*len(b) > MAX_STRING_LENGTH:
-            raise StringTooLong("Sorry, a string that long is not allowed")
-        elif isinstance(b, int) and b*len(a) > MAX_STRING_LENGTH:
-            raise StringTooLong("Sorry, a string that long is not allowed")
+    ''' limit the number of times an iterable can be repeated... '''
+
+    if hasattr(a, '__len__') and b*len(a) > MAX_STRING_LENGTH:
+        raise IterableTooLong('Sorry, I will not evalute something that long.')
+    if hasattr(b, '__len__') and a*len(b) > MAX_STRING_LENGTH:
+        raise IterableTooLong('Sorry, I will not evalute something that long.')
 
     return a * b
 
 
 def safe_add(a, b):  # pylint: disable=invalid-name
-    ''' string length limit again '''
-    if isinstance(a, str) and isinstance(b, str):
+    ''' iterable length limit again '''
+    if hasattr(a, '__len__') and hasattr(b, '__len__'):
         if len(a) + len(b) > MAX_STRING_LENGTH:
-            raise StringTooLong("Sorry, adding those two strings would"
-                                " make a too long string.")
+            raise IterableTooLong("Sorry, adding those two together would"
+                                  " make something too long.")
     return a + b
 
 
@@ -193,11 +195,15 @@ def safe_add(a, b):  # pylint: disable=invalid-name
 # Defaults for the evaluator:
 
 DEFAULT_OPERATORS = {ast.Add: safe_add, ast.Sub: op.sub, ast.Mult: safe_mult,
-                     ast.Div: op.truediv, ast.Pow: safe_power, ast.Mod: op.mod,
+                     ast.Div: op.truediv, ast.FloorDiv: op.floordiv,
+                     ast.Pow: safe_power, ast.Mod: op.mod,
                      ast.Eq: op.eq, ast.NotEq: op.ne,
                      ast.Gt: op.gt, ast.Lt: op.lt,
-                     ast.GtE: op.ge, ast.LtE: op.le, ast.USub: op.neg,
-                     ast.UAdd: op.pos}
+                     ast.GtE: op.ge, ast.LtE: op.le,
+                     ast.USub: op.neg, ast.UAdd: op.pos,
+                     ast.In: lambda x, y: op.contains(y, x),
+                     ast.NotIn: lambda x, y: not op.contains(y, x),
+                     }
 
 DEFAULT_FUNCTIONS = {"rand": random, "randint": random_int,
                      "int": int, "float": float,
@@ -233,6 +239,28 @@ class SimpleEval(object):  # pylint: disable=too-few-public-methods
         self.functions = functions
         self.names = names
 
+        self.nodes = {
+            ast.Num: self._eval_num,
+            ast.Str: self._eval_str,
+            ast.Name: self._eval_name,
+            ast.UnaryOp: self._eval_unaryop,
+            ast.BinOp: self._eval_binop,
+            ast.BoolOp: self._eval_boolop,
+            ast.Compare: self._eval_compare,
+            ast.IfExp: self._eval_ifexp,
+            ast.Call: self._eval_call,
+            ast.keyword: self._eval_keyword,
+            ast.Name: self._eval_name,
+            ast.Subscript: self._eval_subscript,
+            ast.Attribute: self._eval_attribute,
+            ast.Index: self._eval_index,
+            ast.Slice: self._eval_slice,
+        }
+
+        # py3k stuff:
+        if hasattr(ast, 'NameConstant'):
+            self.nodes[ast.NameConstant] = self._eval_nameconstant
+
     def eval(self, expr):
         ''' evaluate an expresssion, using the operators, functions and
             names previously set up. '''
@@ -244,135 +272,184 @@ class SimpleEval(object):  # pylint: disable=too-few-public-methods
         # and evaluate:
         return self._eval(ast.parse(expr.strip()).body[0].value)
 
-    # pylint: disable=too-many-return-statements, too-many-branches
     def _eval(self, node):
         ''' The internal evaluator used on each node in the parsed tree. '''
 
-        # literals:
+        try:
+            handler = self.nodes[type(node)]
+        except KeyError:
+            raise FeatureNotAvailable("Sorry, {0} is not available in this "
+                                      "evaluator".format(type(node).__name__))
 
-        if isinstance(node, ast.Num):  # <number>
-            return node.n
-        elif isinstance(node, ast.Str):  # <string>
-            if len(node.s) > MAX_STRING_LENGTH:
-                raise StringTooLong("String Literal in statement is too long!"
-                                    " ({0}, when {1} is max)".format(
-                                        len(node.s), MAX_STRING_LENGTH))
-            return node.s
+        return handler(node)
 
-        # python 3 compatibility:
+    def _eval_num(self, node):
+        return node.n
 
-        elif (hasattr(ast, 'NameConstant') and
-              isinstance(node, ast.NameConstant)):  # <bool>
-            return node.value
+    def _eval_str(self, node):
+        if len(node.s) > MAX_STRING_LENGTH:
+            raise IterableTooLong("String Literal in statement is too long!"
+                                  " ({0}, when {1} is max)".format(
+                                    len(node.s), MAX_STRING_LENGTH))
+        return node.s
 
-        # operators, functions, etc:
+    def _eval_nameconstant(self, node):
+        return node.value
 
-        elif isinstance(node, ast.UnaryOp):  # - and + etc.
-            return self.operators[type(node.op)](self._eval(node.operand))
-        elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
-            return self.operators[type(node.op)](self._eval(node.left),
-                                                 self._eval(node.right))
-        elif isinstance(node, ast.BoolOp):  # and & or...
-            if isinstance(node.op, ast.And):
-                for value in node.values:
-                    vout = self._eval(value)
-                    if not vout:
-                        return False
-                return vout
-            elif isinstance(node.op, ast.Or):
-                for value in node.values:
-                    vout = self._eval(value)
-                    if vout:
-                        return vout
-                return False
+    def _eval_unaryop(self, node):
+        return self.operators[type(node.op)](self._eval(node.operand))
 
-        elif isinstance(node, ast.Compare):  # 1 < 2, a == b...
-            left = self._eval(node.left)
-            for operation, comp in zip(node.ops, node.comparators):
-                right = self._eval(comp)
-                if self.operators[type(operation)](left, right):
-                    left = right  # Hi Dr. Seuss...
-                else:
+    def _eval_binop(self, node):
+        return self.operators[type(node.op)](self._eval(node.left),
+                                             self._eval(node.right))
+
+    def _eval_boolop(self, node):
+        if isinstance(node.op, ast.And):
+            for value in node.values:
+                vout = self._eval(value)
+                if not vout:
                     return False
-            return True
+            return vout
+        elif isinstance(node.op, ast.Or):
+            for value in node.values:
+                vout = self._eval(value)
+                if vout:
+                    return vout
+            return False
 
-        elif isinstance(node, ast.IfExp):  # x if y else z
-            return self._eval(node.body) if self._eval(node.test) \
+    def _eval_compare(self, node):
+        left = self._eval(node.left)
+        for operation, comp in zip(node.ops, node.comparators):
+            right = self._eval(comp)
+            if self.operators[type(operation)](left, right):
+                left = right  # Hi Dr. Seuss...
+            else:
+                return False
+        return True
+
+    def _eval_ifexp(self, node):
+        return self._eval(node.body) if self._eval(node.test) \
                                          else self._eval(node.orelse)
-        elif isinstance(node, ast.Call):  # function...
+
+    def _eval_call(self, node):
+        if isinstance(node.func, ast.Attribute):
+            func = self._eval(node.func)
+        else:
             try:
-                if isinstance(node.func, ast.Name):
-                    return self.functions[node.func.id](*(self._eval(a)
-                                                          for a in node.args))
-                elif isinstance(node.func, ast.Attribute):
-                    return self._eval(node.func)(*(self._eval(a)
-                                                   for a in node.args))
+                func = self.functions[node.func.id]
             except KeyError:
                 raise FunctionNotDefined(node.func.id, self.expr)
 
-        # variables/names:
+        return func(
+            *(self._eval(a) for a in node.args),
+            **dict(self._eval(k) for k in node.keywords)
+        )
 
-        elif isinstance(node, ast.Name):  # a, b, c...
-            try:
-                # This happens at least for slicing
-                # This is a safe thing to do because it is impossible
-                # that there is a true exression assigning to none
-                # (the compiler rejects it, so you can't even pass that
-                # to ast.parse)
-                if node.id == "None":
-                    return None
-                elif isinstance(self.names, dict):
-                    return self.names[node.id]
-                elif callable(self.names):
-                    return self.names(node)
-                else:
-                    raise InvalidExpression(
-                        'Trying to use name (variable) "{0}"'
-                        ' when no "names" defined for'
-                        ' evaluator'.format(node.id))
+    def _eval_keyword(self, node):
+        return node.arg, self._eval(node.value)
 
-            except KeyError:
-                raise NameNotDefined(node.id, self.expr)
+    def _eval_name(self, node):
+        try:
+            # This happens at least for slicing
+            # This is a safe thing to do because it is impossible
+            # that there is a true exression assigning to none
+            # (the compiler rejects it, so you can't even
+            # pass that to ast.parse)
+            if node.id == "None":
+                return None
+            elif isinstance(self.names, dict):
+                return self.names[node.id]
+            elif callable(self.names):
+                return self.names(node)
+            else:
+                raise InvalidExpression('Trying to use name (variable) "{0}"'
+                                        ' when no "names" defined for'
+                                        ' evaluator'.format(node.id))
 
-        elif isinstance(node, ast.Subscript):  # b[1]
-            return self._eval(node.value)[self._eval(node.slice)]
+        except KeyError:
+            raise NameNotDefined(node.id, self.expr)
 
-        elif isinstance(node, ast.Attribute):  # a.b.c
-            for prefix in DISALLOW_PREFIXES:
-                if node.attr.startswith(prefix):
-                    raise FeatureNotAvailable(
-                        "Sorry, access to __attributes "
-                        " or func_ attributes is not available. "
-                        "({0})".format(node.attr))
+    def _eval_subscript(self, node):
 
-            try:
-                return self._eval(node.value)[node.attr]
-            except (KeyError, TypeError):
-                pass
+        container = self._eval(node.value)
+        key = self._eval(node.slice)
+        try:
+            return container[key]
+        except KeyError:
+            raise
 
-            # Maybe the base object is an actual object, not just a dict
-            try:
-                return getattr(self._eval(node.value), node.attr)
-            except (AttributeError, TypeError):
-                pass
+        return self._eval(node.value)[self._eval(node.slice)]
 
-            # If it is neither, raise an exception
-            raise AttributeDoesNotExist(node.attr, self.expr)
+    def _eval_attribute(self, node):
+        for prefix in DISALLOW_PREFIXES:
+            if node.attr.startswith(prefix):
+                raise FeatureNotAvailable(
+                    "Sorry, access to __attributes "
+                    " or func_ attributes is not available. "
+                    "({0})".format(node.attr))
 
-        elif isinstance(node, ast.Index):
-            return self._eval(node.value)
-        elif isinstance(node, ast.Slice):
-            lower = upper = step = None
-            if node.lower is not None:
-                lower = self._eval(node.lower)
-            if node.upper is not None:
-                upper = self._eval(node.upper)
-            if node.step is not None:
-                step = self._eval(node.step)
-            return slice(lower, upper, step)
-        else:
-            raise FeatureNotAvailable("Sorry, {0} is not available in this "
-                                      "evaluator".format(type(node).__name__))
+        try:
+            return self._eval(node.value)[node.attr]
+        except (KeyError, TypeError):
+            pass
+
+        # Maybe the base object is an actual object, not just a dict
+        try:
+            return getattr(self._eval(node.value), node.attr)
+        except (AttributeError, TypeError):
+            pass
+
+        # If it is neither, raise an exception
+        raise AttributeDoesNotExist(node.attr, self.expr)
+
+    def _eval_index(self, node):
+        return self._eval(node.value)
+
+    def _eval_slice(self, node):
+        lower = upper = step = None
+        if node.lower is not None:
+            lower = self._eval(node.lower)
+        if node.upper is not None:
+            upper = self._eval(node.upper)
+        if node.step is not None:
+            step = self._eval(node.step)
+        return slice(lower, upper, step)
+
+
+class EvalWithCompoundTypes(SimpleEval):
+    '''
+        SimpleEval with additional Compound Types, and their respective
+        function editions. (list, tuple, dict, set).
+    '''
+
+    def __init__(self, *args, **kwargs):
+        super(EvalWithCompoundTypes, self).__init__(*args, **kwargs)
+
+        self.functions.update(
+            list=list,
+            tuple=tuple,
+            dict=dict,
+            set=set)
+
+        self.nodes.update({
+            ast.Dict: self._eval_dict,
+            ast.Tuple: self._eval_tuple,
+            ast.List: self._eval_list,
+            ast.Set: self._eval_set
+        })
+
+    def _eval_dict(self, node):
+        return {self._eval(k): self._eval(v)
+                for (k, v) in zip(node.keys, node.values)}
+
+    def _eval_tuple(self, node):
+        return tuple(self._eval(x) for x in node.elts)
+
+    def _eval_list(self, node):
+        return list(self._eval(x) for x in node.elts)
+
+    def _eval_set(self, node):
+        return set(self._eval(x) for x in node.elts)
 
 
 def simple_eval(expr, operators=None, functions=None, names=None):
